"""This module contains the code to get forecast from WeatherFlows Better Forecast API."""

from __future__ import annotations

import abc
import datetime
import json
import logging

from typing import Any
from urllib.request import urlopen

import aiohttp

from .const import (
    ICON_LIST,
    WEATHERFLOW_DEVICE_URL,
    WEATHERFLOW_FORECAST_URL,
    WEATHERFLOW_SENSOR_URL,
    WEATHERFLOW_STATION_URL,
)
from .data import (
    WeatherFlowDeviceData,
    WeatherFlowForecastData,
    WeatherFlowForecastDaily,
    WeatherFlowForecastHourly,
    WeatherFlowSensorData,
    WeatherFlowStationData,
)

_LOGGER = logging.getLogger(__name__)

class WeatherFlowForecastBadRequest(Exception):
    """Request is invalid."""


class WeatherFlowForecastUnauthorized(Exception):
    """Unauthorized API Key."""


class WeatherFlowForecastWongStationId(Exception):
    """Station ID does not exist."""


class WeatherFlowForecastInternalServerError(Exception):
    """Servers encounter an unexpected error."""


class WeatherFlowAPIBase:
    """Baseclass to use as dependency injection pattern for easier automatic testing."""

    @abc.abstractmethod
    def api_request( self, url: str) -> dict[str, Any]:
        """Override this."""
        raise NotImplementedError(
            "users must define api_request to use this base class"
        )

    @abc.abstractmethod
    async def async_api_request( self, url: str) -> dict[str, Any]:
        """Override this."""
        raise NotImplementedError(
            "users must define async_api_request to use this base class"
        )

class WeatherFlowAPI(WeatherFlowAPIBase):
    """Default implementation for WeatherFlow api."""

    def __init__(self) -> None:
        """Init the API with or without session."""
        self.session = None

    def api_request(self, url: str) -> dict[str, Any]:
        """Return data from API."""
        _LOGGER.debug("URL: %s", url)

        response = urlopen(url)
        data = response.read().decode("utf-8")
        json_data = json.loads(data)

        return json_data

    async def async_api_request(self, url: str) -> dict[str, Any]:
        """Get data from WeatherFlow API."""

        _LOGGER.debug("URL CALLED: %s", url)

        is_new_session = False
        if self.session is None:
            self.session = aiohttp.ClientSession()
            is_new_session = True

        async with self.session.get(url) as response:
            if response.status != 200:
                if is_new_session:
                    await self.session.close()
                if response.status == 400:
                    raise WeatherFlowForecastBadRequest(
                        "400 BAD_REQUEST: Requests is invalid in some way (invalid dates, bad location parameter etc)."
                    )
                if response.status == 401:
                    raise WeatherFlowForecastUnauthorized(
                        "401 UNAUTHORIZED: The API token is incorrect or your account status is inactive or disabled."
                    )
                if response.status == 404:
                    raise WeatherFlowForecastWongStationId(
                        "404 NOT FOUND: The ID of the Station or Device cannot be found."
                    )
                if response.status == 500:
                    raise WeatherFlowForecastInternalServerError(
                        "500 INTERNAL_SERVER_ERROR: WeatherFlow servers encounter an unexpected error."
                    )

            data = await response.text()
            if is_new_session:
                await self.session.close()

            json_data = json.loads(data)

            return json_data


class WeatherFlow:
    """Class that uses the Better Forecast API from WeatherFlow to retreive forecast data."""

    def __init__(
        self,
        station_id: int,
        api_token: str,
        session: aiohttp.ClientSession = None,
        elevation = None,
        api: WeatherFlowAPIBase = WeatherFlowAPI(),
    ) -> None:
        """Return data from WeatherFlow API."""
        self._station_id = station_id
        self._api_token = api_token
        self._elevation = elevation
        self._api = api
        self._device_id = None
        self._tempest_device = False
        self._json_data = None
        self._station_data: WeatherFlowStationData = None
        self._device_data: WeatherFlowDeviceData = None
        self._voltage: float = None

        if session:
            self._api.session = session


    def get_forecast(self) -> list[WeatherFlowForecastData]:
        """Return list of forecasts. The first in list are the current one."""
        api_url = f"{WEATHERFLOW_FORECAST_URL}{self._station_id}&token={self._api_token}"
        self._json_data = self._api.api_request(api_url)

        return _get_forecast(self._json_data)

    def get_station(self) -> list[WeatherFlowStationData]:
        """Return list of station information."""
        station_url = f"{WEATHERFLOW_STATION_URL}{self._station_id}?token={self._api_token}"
        json_data = self._api.api_request(station_url)

        return _get_station(json_data)

    def fetch_sensor_data(self, voltage: float = None) -> list[WeatherFlowSensorData]:
        """Return list of sensor data."""
        device_data = None
        sensor_data = None

        if self._device_id is None and not self._tempest_device:
            station_url = f"{WEATHERFLOW_STATION_URL}{self._station_id}?token={self._api_token}"
            json_station_data = self._api.api_request(station_url)
            station_data: WeatherFlowStationData = _get_station(json_station_data)
            self._device_id = station_data.device_id
            self._tempest_device = False if self._device_id is None else True

        if self._device_id is not None:
            _device_id = station_data.device_id
            device_url = f"{WEATHERFLOW_DEVICE_URL}{_device_id}?token={self._api_token}"
            json_device_data = self._api.api_request(device_url)
            device_data: WeatherFlowDeviceData = _get_device_data(json_device_data, _device_id)

        if device_data is not None or not self._tempest_device:
            _precipitation_type = device_data.precipitation_type if self._tempest_device else None
            _voltage = device_data.voltage if self._tempest_device else None
            api_url = f"{WEATHERFLOW_SENSOR_URL}{self._station_id}?token={self._api_token}"
            json_data = self._api.api_request(api_url)
            sensor_data = _get_sensor_data(json_data, self._elevation, _voltage, _precipitation_type, self._station_name)

        return sensor_data

    async def async_fetch_sensor_data(self) -> list[WeatherFlowSensorData]:
        """Return sensor data from API."""
        device_data = None
        sensor_data = None

        if self._device_id is None and not self._tempest_device:
            station_url = f"{WEATHERFLOW_STATION_URL}{self._station_id}?token={self._api_token}"
            json_station_data = await self._api.async_api_request(station_url)
            station_data: WeatherFlowStationData = _get_station(json_station_data)
            self._device_id = station_data.device_id
            self._station_name = station_data.station_name
            self._tempest_device = False if self._device_id is None else True

        if self._device_id is not None:
            device_url = f"{WEATHERFLOW_DEVICE_URL}{self._device_id}?token={self._api_token}"
            json_device_data = await self._api.async_api_request(device_url)
            device_data: WeatherFlowDeviceData = _get_device_data(json_device_data, self._device_id)

        if device_data is not None or not self._tempest_device:
            _precipitation_type = device_data.precipitation_type if self._tempest_device else None
            _voltage = device_data.voltage if self._tempest_device else None
            api_url = f"{WEATHERFLOW_SENSOR_URL}{self._station_id}?token={self._api_token}"
            json_data = await self._api.async_api_request(api_url)
            sensor_data = _get_sensor_data(json_data, self._elevation, _voltage, _precipitation_type, self._station_name)

        return sensor_data

    async def async_get_forecast(self) -> list[WeatherFlowForecastData]:
        """Return list of forecasts. The first in list are the current one."""
        api_url = f"{WEATHERFLOW_FORECAST_URL}{self._station_id}&token={self._api_token}"
        self._json_data = await self._api.async_api_request(api_url)

        return _get_forecast(self._json_data)

    async def async_get_station(self) -> list[WeatherFlowStationData]:
        """Return list with Station information."""
        api_url = f"{WEATHERFLOW_STATION_URL}{self._station_id}?token={self._api_token}"

        json_data = await self._api.async_api_request(api_url)
        station_data = _get_station(json_data)
        self._station_data = station_data
        return station_data


def _calced_day_values(day_number, hourly_data) -> dict[str, Any]:
    """Calculate values for day by using hourly data."""
    _precipitation: float = 0
    _wind_speed = []
    _wind_bearing = []
    _wind_gust = []

    for item in hourly_data:
        if item.get("local_day") == day_number:
            _precipitation += item.get("precip", 0)
            _wind_bearing.append(item.get("wind_direction", 0))
            _wind_speed.append(item.get("wind_avg", 0))
            _wind_gust.append(item.get("wind_gust", 0))

    _sum_wind_speed = sum(_wind_speed) / len(_wind_speed)
    _sum_wind_bearing = sum(_wind_bearing) / len(_wind_bearing)
    _max_wind_gust = max(_wind_gust)

    return {
        "precipitation": _precipitation,
        "wind_bearing": _sum_wind_bearing,
        "wind_speed": _sum_wind_speed,
        "wind_gust": _max_wind_gust,
    }


# pylint: disable=R0914, R0912, W0212, R0915
def _get_forecast(api_result: dict) -> list[WeatherFlowForecastData]:
    """Return WeatherFlowForecast list from API."""

    # Get Current Conditions
    current_conditions: WeatherFlowForecastData = _get_forecast_current(api_result)

    forecasts_daily = []
    forecasts_hourly = []

    # Add daily forecast details
    for item in api_result["forecast"]["daily"]:
        timestamp = item["day_start_local"]
        valid_time = datetime.datetime.fromtimestamp(timestamp)
        condition = item.get("conditions", "Data Error")
        icon = ICON_LIST.get(item["icon"], "exceptional")
        temperature = item.get("air_temp_high", None)
        temp_low = item.get("air_temp_low", None)
        precipitation_probability = item.get("precip_probability", None)
        _calc_values = _calced_day_values(item["day_num"], api_result["forecast"]["hourly"])
        precipitation = _calc_values["precipitation"]
        wind_bearing = _calc_values["wind_bearing"]
        wind_speed = _calc_values["wind_speed"]
        wind_gust = _calc_values["wind_gust"]

        forecast = WeatherFlowForecastDaily(
            valid_time,
            timestamp,
            temperature,
            temp_low,
            condition,
            icon,
            precipitation_probability,
            precipitation,
            wind_bearing,
            wind_speed,
            wind_gust,
        )
        forecasts_daily.append(forecast)

    current_conditions.forecast_daily = forecasts_daily

    # Add Hourly Forecast
    for item in api_result["forecast"]["hourly"]:
        timestamp = item["time"]
        valid_time = datetime.datetime.fromtimestamp(timestamp)
        condition = item.get("conditions", None)
        icon = ICON_LIST.get(item["icon"], "exceptional")
        temperature = item.get("air_temperature", None)
        apparent_temperature = item.get("feels_like", None)
        precipitation = item.get("precip", None)
        precipitation_probability = item.get("precip_probability", None)
        humidity = item.get("relative_humidity", None)
        pressure = item.get("sea_level_pressure", None)
        uv_index = item.get("uv", None)
        wind_speed = item.get("wind_avg", None)
        wind_gust_speed = item.get("wind_gust", None)
        wind_bearing = item.get("wind_direction", None)

        forecast = WeatherFlowForecastHourly(
            valid_time,
            timestamp,
            temperature,
            apparent_temperature,
            condition,
            icon,
            humidity,
            precipitation,
            precipitation_probability,
            pressure,
            wind_bearing,
            wind_gust_speed,
            wind_speed,
            uv_index,
        )
        forecasts_hourly.append(forecast)

    current_conditions.forecast_hourly = forecasts_hourly

    return current_conditions


# pylint: disable=R0914, R0912, W0212, R0915
def _get_forecast_current(api_result: dict) -> list[WeatherFlowForecastData]:
    """Return WeatherFlowForecast list from API."""

    item = api_result["current_conditions"]
    timestamp = item["time"]
    valid_time = datetime.datetime.fromtimestamp(timestamp)
    condition = item.get("conditions", None)
    icon = ICON_LIST.get(item["icon"], "exceptional")
    temperature = item.get("air_temperature", None)
    dew_point = item.get("dew_point", None)
    apparent_temperature = item.get("feels_like", None)
    precipitation = item.get("precip_accum_local_day", None)
    humidity = item.get("relative_humidity", None)
    pressure = item.get("sea_level_pressure", None)
    uv_index = item.get("uv", None)
    wind_speed = item.get("wind_avg", None)
    wind_gust_speed = item.get("wind_gust", None)
    wind_bearing = item.get("wind_direction", None)

    current_condition = WeatherFlowForecastData(
        valid_time,
        timestamp,
        apparent_temperature,
        condition,
        dew_point,
        humidity,
        icon,
        precipitation,
        pressure,
        temperature,
        uv_index,
        wind_bearing,
        wind_gust_speed,
        wind_speed,
    )

    return current_condition


# pylint: disable=R0914, R0912, W0212, R0915
def _get_station(api_result: dict) -> list[WeatherFlowStationData]:
    """Return WeatherFlowForecast list from API."""

    item = api_result["stations"][0]

    station_name = item.get("name", None)
    latitude = item.get("latitude", None)
    longitude = item.get("longitude", None)
    timezone = item.get("timezone", None)
    device_id = None
    firmware_revision = None
    serial_number = None
    for device in item["devices"]:
        if device.get("device_type", None) == "ST":
            device_id = device.get("device_id", None)
            firmware_revision = device.get("firmware_revision", None)
            serial_number = device.get("serial_number", None)
            break

    station_data = WeatherFlowStationData(
        station_name,
        latitude,
        longitude,
        timezone,
        device_id,
        firmware_revision,
        serial_number,
    )

    return station_data


# pylint: disable=R0914, R0912, W0212, R0915
def _get_sensor_data(api_result: dict, elevation: float, voltage: float, precipitation_type: int, station_name: str) -> list[WeatherFlowSensorData]:
    """Return WeatherFlowSensorData list from API."""

    _LOGGER.debug("ELEVATION: %s", elevation)

    item = api_result["obs"][0]

    air_density = item.get("air_density", None)
    air_temperature = item.get("air_temperature", None)
    barometric_pressure = item.get("barometric_pressure", None)
    brightness = item.get("brightness", None)
    delta_t = item.get("delta_t", None)
    dew_point = item.get("dew_point", None)
    feels_like = item.get("feels_like", None)
    heat_index = item.get("heat_index", None)
    lightning_strike_count = item.get("lightning_strike_count", None)
    lightning_strike_count_last_1hr = item.get("lightning_strike_count_last_1hr", None)
    lightning_strike_count_last_3hr = item.get("lightning_strike_count_last_3hr", None)
    lightning_strike_last_distance = item.get("lightning_strike_last_distance", None)
    lightning_strike_last_epoch = item.get("lightning_strike_last_epoch", None)
    precip = item.get("precip", None)
    precip_accum_last_1hr = item.get("precip_accum_last_1hr", None)
    precip_accum_local_day = item.get("precip_accum_local_day", None)
    precip_accum_local_yesterday = item.get("precip_accum_local_yesterday", None)
    precip_minutes_local_day = item.get("precip_minutes_local_day", None)
    precip_minutes_local_yesterday = item.get("precip_minutes_local_yesterday", None)
    pressure_trend = item.get("pressure_trend", None)
    relative_humidity = item.get("relative_humidity", None)
    sea_level_pressure = item.get("sea_level_pressure", None)
    solar_radiation = item.get("solar_radiation", None)
    station_pressure = item.get("station_pressure", None)
    timestamp = item.get("timestamp", None)
    uv = item.get("uv", None)
    wet_bulb_globe_temperature = item.get("wet_bulb_globe_temperature", None)
    wet_bulb_temperature = item.get("wet_bulb_temperature", None)
    wind_avg = item.get("wind_avg", None)
    wind_chill = item.get("wind_chill", None)
    wind_direction = item.get("wind_direction", None)
    wind_gust = item.get("wind_gust", None)
    wind_lull = item.get("wind_lull", None)
    precip_accum_local_day_final = item.get("precip_accum_local_day_final", None)
    precip_accum_local_yesterday_final = item.get("precip_accum_local_yesterday_final", None)
    precip_minutes_local_day_final = item.get("precip_minutes_local_day_final", None)
    precip_minutes_local_yesterday_final = item.get("precip_minutes_local_yesterday_final", None)

    sensor_data = WeatherFlowSensorData(
        air_density,
        air_temperature,
        barometric_pressure,
        brightness,
        delta_t,
        dew_point,
        feels_like,
        heat_index,
        lightning_strike_count,
        lightning_strike_count_last_1hr,
        lightning_strike_count_last_3hr,
        lightning_strike_last_distance,
        lightning_strike_last_epoch,
        precip,
        precip_accum_last_1hr,
        precip_accum_local_day,
        precip_accum_local_yesterday,
        precip_minutes_local_day,
        precip_minutes_local_yesterday,
        precipitation_type,
        pressure_trend,
        relative_humidity,
        sea_level_pressure,
        solar_radiation,
        station_pressure,
        timestamp,
        uv,
        voltage,
        wet_bulb_globe_temperature,
        wet_bulb_temperature,
        wind_avg,
        wind_chill,
        wind_direction,
        wind_gust,
        wind_lull,
        precip_accum_local_day_final,
        precip_accum_local_yesterday_final,
        precip_minutes_local_day_final,
        precip_minutes_local_yesterday_final,
        elevation,
        station_name,
    )

    return sensor_data

# pylint: disable=R0914, R0912, W0212, R0915
def _get_device_data(api_result: dict, device_id: int) -> float:
    """Return WeatherFlow Device Voltage from API."""

    precipitation_type = None if api_result is None else api_result["obs"][0][13]
    voltage = None if api_result is None else api_result["obs"][0][16]

    device_data = WeatherFlowDeviceData(
        device_id,
        voltage,
        precipitation_type,
    )

    return device_data




# ruff: noqa: F401
"""Python Wrapper for WeatherFlow Forecast API."""
from __future__ import annotations
from pyweatherflow_forecast.wffcst_lib import (
    WeatherFlow,
    WeatherFlowForecastInternalServerError,
    WeatherFlowForecastBadRequest,
    WeatherFlowForecastUnauthorized,
    WeatherFlowForecastWongStationId,
)
from pyweatherflow_forecast.data import (
    WeatherFlowForecastData,
    WeatherFlowForecastDaily,
    WeatherFlowDeviceData,
    WeatherFlowForecastHourly,
    WeatherFlowSensorData,
    WeatherFlowStationData,
)

__title__ = "pyweatherflow_forecast"
__version__ = "1.0.6"
__author__ = "briis"
__license__ = "MIT"



"""Holds the Data Calsses for WeatherFlow Forecast Wrapper."""

from __future__ import annotations
from datetime import datetime
import math
import logging
_LOGGER = logging.getLogger(__name__)

class WeatherFlowForecastData:
    """Class to hold forecast data."""

        # pylint: disable=R0913, R0902, R0914
    def __init__(
        self,
        datetime: datetime,
        timestamp: int,
        apparent_temperature: float,
        condition: str,
        dew_point: float,
        humidity: int,
        icon: str,
        precipitation: float,
        pressure: float,
        temperature: float,
        uv_index: int,
        wind_bearing: int,
        wind_gust_speed: float,
        wind_speed: float,
        forecast_daily: WeatherFlowForecastDaily = None,
        forecast_hourly: WeatherFlowForecastHourly = None,
    ) -> None:
        """Dataset constructor."""
        self._datetime = datetime
        self._timestamp = timestamp
        self._apparent_temperature = apparent_temperature
        self._condition = condition
        self._dew_point = dew_point
        self._humidity = humidity
        self._icon = icon
        self._precipitation = precipitation
        self._pressure = pressure
        self._temperature = temperature
        self._uv_index = uv_index
        self._wind_bearing = wind_bearing
        self._wind_gust_speed = wind_gust_speed
        self._wind_speed = wind_speed
        self._forecast_daily = forecast_daily
        self._forecast_hourly = forecast_hourly


    @property
    def temperature(self) -> float:
        """Air temperature (Celcius)."""
        return self._temperature

    @property
    def dew_point(self) -> float:
        """Dew Point (Celcius)."""
        return self._dew_point

    @property
    def condition(self) -> str:
        """Weather condition text."""
        return self._condition

    @property
    def icon(self) -> str:
        """Weather condition symbol."""
        return self._icon

    @property
    def humidity(self) -> int:
        """Humidity (%)."""
        return self._humidity

    @property
    def apparent_temperature(self) -> float:
        """Feels like temperature (Celcius)."""
        return self._apparent_temperature

    @property
    def precipitation(self) -> float:
        """Precipitation (mm)."""
        return self._precipitation

    @property
    def pressure(self) -> float:
        """Sea Level Pressure (MB)."""
        return self._pressure

    @property
    def wind_bearing(self) -> float:
        """Wind bearing (degrees)."""
        return self._wind_bearing

    @property
    def wind_gust_speed(self) -> float:
        """Wind gust (m/s)."""
        return self._wind_gust_speed

    @property
    def wind_speed(self) -> float:
        """Wind speed (m/s)."""
        return self._wind_speed

    @property
    def uv_index(self) -> float:
        """UV Index."""
        return self._uv_index

    @property
    def datetime(self) -> datetime:
        """Valid time."""
        return self._datetime

    @property
    def datetimestamptime(self) -> int:
        """Timestamp."""
        return self._timestamp

    @property
    def forecast_daily(self) -> WeatherFlowForecastDaily:
        """Forecast List."""
        return self._forecast_daily

    @forecast_daily.setter
    def forecast_daily(self, new_forecast):
        """Forecast daily new value."""
        self._forecast_daily = new_forecast

    @property
    def forecast_hourly(self) -> WeatherFlowForecastHourly:
        """Forecast List."""
        return self._forecast_hourly

    @forecast_hourly.setter
    def forecast_hourly(self, new_forecast):
        """Forecast hourly new value."""
        self._forecast_hourly = new_forecast

class WeatherFlowForecastDaily:
    """Class to hold daily forecast data."""

        # pylint: disable=R0913, R0902, R0914
    def __init__(
        self,
        datetime: datetime,
        timestamp: int,
        temperature: float,
        temp_low: float,
        condition: str,
        icon: str,
        precipitation_probability: int,
        precipitation: float,
        wind_bearing: int,
        wind_speed: float,
        wind_gust: float,
    ) -> None:
        """Dataset constructor."""
        self._datetime = datetime
        self._timestamp = timestamp
        self._temperature = temperature
        self._temp_low = temp_low
        self._condition = condition
        self._icon = icon
        self._precipitation_probability = precipitation_probability
        self._precipitation = precipitation
        self._wind_bearing = wind_bearing
        self._wind_speed = wind_speed
        self._wind_gust = wind_gust

    @property
    def datetime(self) -> datetime:
        """Valid time."""
        return self._datetime

    @property
    def timestamp(self) -> int:
        """Timestamp."""
        return self._timestamp

    @property
    def temperature(self) -> float:
        """Air temperature (Celcius)."""
        return self._temperature

    @property
    def temp_low(self) -> float:
        """Air temperature min during the day (Celcius)."""
        return self._temp_low

    @property
    def condition(self) -> str:
        """Weather condition text."""
        return self._condition

    @property
    def icon(self) -> str:
        """Weather condition symbol."""
        return self._icon

    @property
    def precipitation_probability (self) -> int:
        """Posobility of Precipiation (%)."""
        return self._precipitation_probability

    @property
    def precipitation(self) -> float:
        """Precipitation (mm)."""
        return self._precipitation

    @property
    def wind_bearing(self) -> float:
        """Wind bearing (degrees)."""
        return self._wind_bearing

    @property
    def wind_speed(self) -> float:
        """Wind speed (m/s)."""
        return self._wind_speed

    @property
    def wind_gust(self) -> float:
        """Wind gust (m/s)."""
        return self._wind_gust

class WeatherFlowForecastHourly:
    """Class to hold hourly forecast data."""

        # pylint: disable=R0913, R0902, R0914
    def __init__(
        self,
        datetime: datetime,
        timestamp: int,
        temperature: float,
        apparent_temperature: float,
        condition: str,
        icon: str,
        humidity: int,
        precipitation: float,
        precipitation_probability: int,
        pressure: float,
        wind_bearing: float,
        wind_gust_speed: int,
        wind_speed: int,
        uv_index: float,
    ) -> None:
        """Dataset constructor."""
        self._datetime = datetime
        self._timestamp = timestamp
        self._temperature = temperature
        self._apparent_temperature = apparent_temperature
        self._condition = condition
        self._icon = icon
        self._humidity = humidity
        self._precipitation = precipitation
        self._precipitation_probability = precipitation_probability
        self._pressure = pressure
        self._wind_bearing = wind_bearing
        self._wind_gust_speed = wind_gust_speed
        self._wind_speed = wind_speed
        self._uv_index = uv_index

    @property
    def temperature(self) -> float:
        """Air temperature (Celcius)."""
        return self._temperature

    @property
    def condition(self) -> str:
        """Weather condition text."""
        return self._condition

    @property
    def icon(self) -> str:
        """Weather condition symbol."""
        return self._icon

    @property
    def humidity(self) -> int:
        """Humidity (%)."""
        return self._humidity

    @property
    def apparent_temperature(self) -> float:
        """Feels like temperature (Celcius)."""
        return self._apparent_temperature

    @property
    def precipitation(self) -> float:
        """Precipitation (mm)."""
        return self._precipitation

    @property
    def precipitation_probability (self) -> int:
        """Posobility of Precipiation (%)."""
        return self._precipitation_probability

    @property
    def pressure(self) -> float:
        """Sea Level Pressure (MB)."""
        return self._pressure

    @property
    def wind_bearing(self) -> float:
        """Wind bearing (degrees)."""
        return self._wind_bearing

    @property
    def wind_gust_speed(self) -> float:
        """Wind gust (m/s)."""
        return self._wind_gust_speed

    @property
    def wind_speed(self) -> float:
        """Wind speed (m/s)."""
        return self._wind_speed

    @property
    def uv_index(self) -> float:
        """UV Index."""
        return self._uv_index

    @property
    def datetime(self) -> datetime:
        """Valid time."""
        return self._datetime

    @property
    def timestamp(self) -> int:
        """Timestamp."""
        return self._timestamp


class WeatherFlowDeviceData:
    """Class to hold device data."""

        # pylint: disable=R0913, R0902, R0914
    def __init__(
            self,
            device_id: int,
            voltage: float,
            precipitation_type: int,
    ) -> None:
        """Dataset constructor."""
        self._device_id = device_id
        self._voltage = voltage
        self._precipitation_type = precipitation_type

    @property
    def device_id(self) -> int:
        """Return device id."""
        return self._device_id

    @property
    def voltage(self) -> float:
        """Return voltage of device."""
        return self._voltage

    @property
    def precipitation_type(self) -> int:
        """Return Precipiation type."""
        return self._precipitation_type

    @property
    def battery(self) -> int:
        """Battery (%)."""
        if self._voltage is None:
            return None

        if self._voltage > 2.80:
            _percent = 100
        elif self._voltage < 1.80:
            _percent = 0
        else:
            _percent = (self._voltage - 1.8) * 100

        return _percent

class WeatherFlowStationData:
    """Class to hold station data."""

        # pylint: disable=R0913, R0902, R0914
    def __init__(
            self,
            station_name: str,
            latitude: float,
            longitude: float,
            timezone: str,
            device_id: int,
            firmware_revision: str,
            serial_number: str,
    ) -> None:
        """Dataset constructor."""
        self._station_name = station_name
        self._latitude = latitude
        self._longitude = longitude
        self._timezone = timezone
        self._device_id = device_id
        self._firmware_revision = firmware_revision
        self._serial_number = serial_number

    @property
    def station_name(self) -> str:
        """Name of the Station."""
        return self._station_name

    @property
    def latitude(self) -> float:
        """Latitude of station."""
        return self._latitude

    @property
    def longitude(self) -> float:
        """Longitude of station."""
        return self._longitude

    @property
    def timezone(self) -> str:
        """Timezone of station."""
        return self._timezone

    @property
    def device_id(self) -> int:
        """Device ID."""
        return self._device_id

    @property
    def firmware_revision(self) -> str:
        """Firmware Version."""
        return self._firmware_revision

    @property
    def serial_number(self) -> str:
        """Device Serial Number."""
        return self._serial_number

class WeatherFlowSensorData:
    """Class to hold sensor data."""

        # pylint: disable=R0913, R0902, R0914
    def __init__(
            self,
            air_density: float,
            air_temperature: float,
            barometric_pressure: float,
            brightness: int,
            delta_t: float,
            dew_point: float,
            feels_like: float,
            heat_index: float,
            lightning_strike_count: int,
            lightning_strike_count_last_1hr: int,
            lightning_strike_count_last_3hr: int,
            lightning_strike_last_distance: int,
            lightning_strike_last_epoch: int,
            precip: float,
            precip_accum_last_1hr: float,
            precip_accum_local_day: float,
            precip_accum_local_yesterday: float,
            precip_minutes_local_day: int,
            precip_minutes_local_yesterday: int,
            precipitation_type: int,
            pressure_trend: str,
            relative_humidity: int,
            sea_level_pressure: float,
            solar_radiation: float,
            station_pressure: float,
            timestamp: int,
            uv: float,
            voltage: float,
            wet_bulb_globe_temperature: float,
            wet_bulb_temperature: float,
            wind_avg: float,
            wind_chill: float,
            wind_direction: int,
            wind_gust: float,
            wind_lull: float,
            precip_accum_local_day_final: float,
            precip_accum_local_yesterday_final: float,
            precip_minutes_local_day_final: int,
            precip_minutes_local_yesterday_final: int,
            elevation: float,
            station_name: str,
    ) -> None:
        """Dataset constructor."""
        self._air_density = air_density
        self._air_temperature = air_temperature
        self._barometric_pressure = barometric_pressure
        self._brightness = brightness
        self._delta_t = delta_t
        self._dew_point = dew_point
        self._feels_like = feels_like
        self._heat_index = heat_index
        self._lightning_strike_count = lightning_strike_count
        self._lightning_strike_count_last_1hr = lightning_strike_count_last_1hr
        self._lightning_strike_count_last_3hr = lightning_strike_count_last_3hr
        self._lightning_strike_last_distance = lightning_strike_last_distance
        self._lightning_strike_last_epoch = lightning_strike_last_epoch
        self._precip = precip
        self._precip_accum_last_1hr = precip_accum_last_1hr
        self._precip_accum_local_day = precip_accum_local_day
        self._precip_accum_local_yesterday = precip_accum_local_yesterday
        self._precip_minutes_local_day = precip_minutes_local_day
        self._precip_minutes_local_yesterday = precip_minutes_local_yesterday
        self._precipitation_type = precipitation_type
        self._pressure_trend = pressure_trend
        self._relative_humidity = relative_humidity
        self._sea_level_pressure = sea_level_pressure
        self._solar_radiation = solar_radiation
        self._station_pressure = station_pressure
        self._timestamp = timestamp
        self._uv = uv
        self._voltage = voltage
        self._wet_bulb_globe_temperature = wet_bulb_globe_temperature
        self._wet_bulb_temperature = wet_bulb_temperature
        self._wind_avg = wind_avg
        self._wind_chill = wind_chill
        self._wind_direction = wind_direction
        self._wind_gust = wind_gust
        self._wind_lull = wind_lull
        self._precip_accum_local_day_final = precip_accum_local_day_final
        self._precip_accum_local_yesterday_final = precip_accum_local_yesterday_final
        self._precip_minutes_local_day_final = precip_minutes_local_day_final
        self._precip_minutes_local_yesterday_final = precip_minutes_local_yesterday_final
        self._elevation = elevation
        self._station_name = station_name

    @property
    def absolute_humidity(self) -> float:
        """Aboslute Humidity (g.m-3)."""
        if self._air_temperature is None or self._relative_humidity is None:
            return None

        kelvin = self._air_temperature + 273.16
        humidity = self._relative_humidity / 100
        return (1320.65 / kelvin) * humidity * (10 ** ((7.4475 * (kelvin - 273.14)) / (kelvin - 39.44)))

    @property
    def air_density(self) -> float:
        """Air Density."""
        return self._air_density

    @property
    def air_temperature(self) -> float:
        """Outside Temperature."""
        return self._air_temperature

    @property
    def barometric_pressure(self) -> float:
        """Barometric Pressure."""
        return self._barometric_pressure

    @property
    def battery(self) -> int:
        """Battery (%)."""
        if self._voltage is None:
            return None

        if self._voltage > 2.80:
            _percent = 100
        elif self._voltage < 1.80:
            _percent = 0
        else:
            _percent = (self._voltage - 1.8) * 100

        return _percent

    @property
    def beaufort(self) -> int:
        """Beaufort Value."""
        if self._wind_avg is None:
            return None

        mapping_text = {
            "32.7": 12,
            "28.5": 11,
            "24.5": 10,
            "20.8": 9,
            "17.2": 8,
            "13.9": 7,
            "10.8": 6,
            "8.0": 5,
            "5.5": 4,
            "3.4": 3,
            "1.6": 2,
            "0.3": 1,
            "-1": 0,
        }

        for key, value in mapping_text.items():
            if self._wind_avg > float(key):
                return value
        return None

    @property
    def beaufort_description(self) -> str:
        """Beaufort Textual Description."""

        if self._wind_avg is None:
            return None

        mapping_text = {
            "32.7": "hurricane",
            "28.5": "violent_storm",
            "24.5": "storm",
            "20.8": "strong_gale",
            "17.2": "fresh_gale",
            "13.9": "moderate_gale",
            "10.8": "strong_breeze",
            "8.0": "fresh_breeze",
            "5.5": "moderate_breeze",
            "3.4": "gentle_breeze",
            "1.6": "light_breeze",
            "0.3": "light_air",
            "-1": "calm",
        }

        for key, value in mapping_text.items():
            if self._wind_avg > float(key):
                return value
        return None

    @property
    def brightness(self) -> int:
        """Brightness."""
        return self._brightness

    @property
    def cloud_base(self) -> float:
        """Cloud Base (km)."""
        if self._elevation is None or self._air_temperature is None or self._dew_point is None:
            return None

        return (self._air_temperature - self._dew_point) * 126 + self._elevation

    @property
    def delta_t(self) -> float:
        """Delta_T temperature."""
        return self._delta_t

    @property
    def dew_point(self) -> float:
        """Dew Point."""
        return self._dew_point

    @property
    def feels_like(self) -> float:
        """Apparent temperature."""
        return self._feels_like

    @property
    def freezing_altitude(self) -> float:
        """Freezing Altitude."""
        if self._elevation is None or self._air_temperature is None:
            return None

        _freezing_line = (192 * self._air_temperature) + self._elevation
        return 0 if _freezing_line < 0 else _freezing_line

    @property
    def heat_index(self) -> float:
        """Heat Index."""
        return self._heat_index

    @property
    def is_freezing(self) -> bool:
        """Return if frost outside."""
        if self._air_temperature is None:
            return None
        return self._air_temperature < 0

    @property
    def is_lightning(self) -> bool:
        """Return if lightning strikes."""
        if self._lightning_strike_count is None:
            return None
        return self._lightning_strike_count > 0

    @property
    def is_raining(self) -> bool:
        """Return if raining."""
        if self._precip is None:
            return None
        return self._precip > 0

    @property
    def lightning_strike_count(self) -> int:
        """Ligntning Strike count."""
        return self._lightning_strike_count

    @property
    def lightning_strike_count_last_1hr(self) -> int:
        """Lightning strike count last hour."""
        return self._lightning_strike_count_last_1hr

    @property
    def lightning_strike_count_last_3hr(self) -> int:
        """Lightning strike count last 3 hours."""
        return self._lightning_strike_count_last_3hr

    @property
    def lightning_strike_last_distance(self) -> int:
        """Distance last lif´ghtning strike."""
        return self._lightning_strike_last_distance

    @property
    def lightning_strike_last_epoch(self) -> int:
        """Last lightning strike epoch time."""
        return self._lightning_strike_last_epoch

    @property
    def power_save_mode(self) -> int:
        """Power Save Mode (Tempest devices)."""
        if self._voltage is None or self._solar_radiation is None:
            return None

        _power_save_mode = None
        if self._voltage >= 2.455:
            _power_save_mode = 0
        elif self._voltage <= 2.355:
            _power_save_mode = 3
        elif self._solar_radiation > 100:
            # Assume charging and Voltage is increasing
            if self._voltage >= 2.41:
                _power_save_mode = 1
            elif self._voltage > 2.375:
                _power_save_mode = 2
            else:
                _power_save_mode = 3
        else:
            # Assume discharging and voltage is decreasing
            if self._voltage > 2.415:
                _power_save_mode = 0
            elif self._voltage > 2.39:
                _power_save_mode = 1
            elif self._voltage > 2.355:
                _power_save_mode = 2
            else:
                _power_save_mode = 3

        return _power_save_mode

    @property
    def precip(self) -> float:
        """Precipitation."""
        return self._precip

    @property
    def precip_rate(self) -> float:
        """Precipitation Rate."""
        if self._precip is None:
            return None
        return self._precip * 60

    @property
    def precip_accum_last_1hr(self) -> float:
        """Precipitation last hour."""
        return self._precip_accum_last_1hr

    @property
    def precip_accum_local_day(self) -> float:
        """Prepitation current day."""
        return self._precip_accum_local_day

    @property
    def precip_accum_local_yesterday(self) -> float:
        """Precipitation yesterday."""
        return self._precip_accum_local_yesterday

    @property
    def precip_intensity(self) -> str:
        """Return a string with precipitation intensity."""
        if self._precip is None:
            return None

        mapping_text = {
            "0.01": "no_rain",
            "0.25": "very_light",
            "1": "light",
            "4": "moderate",
            "16": "heavy",
            "50": "very_heavy",
            "1000": "extreme",
        }

        for key, value in mapping_text.items():
            if (self._precip * 60) < float(key):
                return value
        return None

    @property
    def precip_minutes_local_day(self) -> int:
        """Precipitation minutes today."""
        return self._precip_minutes_local_day

    @property
    def precip_minutes_local_yesterday(self) -> int:
        """Precipitation minutes yesterday."""
        return self._precip_minutes_local_yesterday

    @property
    def precip_accum_local_day_final(self) -> float:
        """Prepitation current day (Rain Check)."""
        return self._precip_accum_local_day_final

    @property
    def precip_accum_local_yesterday_final(self) -> float:
        """Precipitation yesterday (Rain Check)."""
        return self._precip_accum_local_yesterday_final

    @property
    def precip_minutes_local_day_final(self) -> int:
        """Precipitation minutes today (Rain Check)."""
        return self._precip_minutes_local_day_final

    @property
    def precip_minutes_local_yesterday_final(self) -> int:
        """Precipitation minutes yesterday (Rain Check)."""
        return self._precip_minutes_local_yesterday_final

    @property
    def precip_type(self) -> str:
        """Return precipitation type."""
        return self._precipitation_type

    @property
    def pressure_trend(self) -> str:
        """Pressure trend text."""
        return self._pressure_trend

    @property
    def relative_humidity(self) -> int:
        """Relative humidity (%)."""
        return self._relative_humidity

    @property
    def sea_level_pressure(self) -> float:
        """Sea level pressure."""
        return self._sea_level_pressure

    @property
    def solar_radiation(self) -> float:
        """Solar Radiation."""
        return self._solar_radiation

    @property
    def station_name(self) -> str:
        """Station name."""
        return self._station_name

    @property
    def station_pressure(self) -> float:
        """Station pressure."""
        return self._station_pressure

    @property
    def timestamp(self) -> int:
        """Time of data update."""
        return self._timestamp

    @property
    def uv(self) -> float:
        """UV index."""
        return self._uv

    @property
    def uv_description(self) -> str:
        """UV value description."""
        if self._uv is None:
            return None

        mapping_text = {
            "10.5": "extreme",
            "7.5": "very-high",
            "5.5": "high",
            "2.8": "moderate",
            "0": "low",
        }

        for key, value in mapping_text.items():
            if self._uv > float(key):
                return value
        return None


    @property
    def visibility(self) -> float:
        """Visibility (km)."""
        if self._elevation is None or self._air_temperature is None or self._relative_humidity is None or self._dew_point is None:
            return None

        _elevation_min = float(2)
        if self._elevation > 2:
            _elevation_min = self._elevation

        _max_visibility = float(3.56972 * math.sqrt(_elevation_min))
        _percent_reduction_a = float((1.13 * abs(self._air_temperature - self._dew_point) - 1.15) / 10)
        if _percent_reduction_a > 1:
            _percent_reduction = float(1)
        elif _percent_reduction_a < 0.025:
            _percent_reduction = 0.025
        else:
            _percent_reduction = _percent_reduction_a

        return float(_max_visibility * _percent_reduction)

    @property
    def voltage(self) -> float:
        """Return voltage of device."""
        return self._voltage

    @property
    def wet_bulb_globe_temperature(self) -> float:
        """Wet bulb globe temperature."""
        return self._wet_bulb_globe_temperature

    @property
    def wet_bulb_temperature(self) -> float:
        """Wet bulb temperature."""
        return self._wet_bulb_temperature

    @property
    def wind_avg(self) -> float:
        """Wind speed."""
        return self._wind_avg

    @property
    def wind_cardinal(self) -> str:
        """Wind Cardinal."""
        if self._wind_direction is None:
            return None

        direction_array = ["n", "nne", "ne", "ene", "e", "ese", "se", "sse", "s", "ssw", "sw", "wsw", "w", "wnw", "nw", "nnw", "n"]
        return direction_array[int((self._wind_direction + 11.25) / 22.5)]

    @property
    def wind_chill(self) -> float:
        """Wind chill factor."""
        return self._wind_chill

    @property
    def wind_direction(self) -> int:
        """Wind direction in degrees."""
        return self._wind_direction

    @property
    def wind_gust(self) -> float:
        """Wind gust speed."""
        return self._wind_gust

    @property
    def wind_lull(self) -> float:
        """Wind lull speed."""
        return self._wind_lull



"""System Wide constants for WeatherFlow Forecast Wrapper."""
from __future__ import annotations

CACHE_MINUTES = 30

FORECAST_TYPE_DAILY = 0
FORECAST_TYPE_HOURLY = 1

WEATHERFLOW_BASE_URL = "https://swd.weatherflow.com/swd/rest"
WEATHERFLOW_DEVICE_URL = f"{WEATHERFLOW_BASE_URL}/observations/device/"
WEATHERFLOW_FORECAST_URL = f"{WEATHERFLOW_BASE_URL}/better_forecast?station_id="
WEATHERFLOW_SENSOR_URL = f"{WEATHERFLOW_BASE_URL}/observations/station/"
WEATHERFLOW_STATION_URL = f"{WEATHERFLOW_BASE_URL}/stations/"

ICON_LIST = {
    "clear-day": "sunny",
    "clear-night": "clear-night",
    "cloudy": "cloudy",
    "foggy": "fog",
    "partly-cloudy-day": "partlycloudy",
    "partly-cloudy-night": "partlycloudy",
    "possibly-rainy-day": "rainy",
    "possibly-rainy-night": "rainy",
    "possibly-sleet-day": "snowy-rainy",
    "possibly-sleet-night": "snowy-rainy",
    "possibly-snow-day": "snowy",
    "possibly-snow-night": "snowy",
    "possibly-thunderstorm-day": "lightning-rainy",
    "possibly-thunderstorm-night": "lightning-rainy",
    "rainy": "rainy",
    "sleet": "snowy-rainy",
    "snow": "snowy",
    "thunderstorm": "lightning",
    "windy": "windy"
}
